open TestDriver

(* Definitions and specifications for numbers of all kinds *)


(******************************************************************************)
(** Addition                                                                  *)
(******************************************************************************)

(* we've split the Addable module type into a core interface and an extension
   interface.  The Core interface defines the bare minimum that
   must be implemented (by the guy in the "private hat").  The
   Extension interface contains helper methods that are useful for the user, but
   can be implemented in terms of the Core interface

   Ints, Floats, Vectors, Matrices, Functions, etc. are addable.
   Strings are not.

   Implementations of AddableCore should satisfy the Addable spec, documented
   below
*)

module type AddableCore = sig
  type t

  val (+)   : t -> t -> t

  (* ocaml uses ~- to implement unary negation *)
  val (~-)  : t -> t

  (* these are required to give good specs for + and - *)
  val zero : t
  val (=)  : t -> t -> bool
end

module AddableSpec (C : AddableCore)
                   (E : Examples with type t := C.t) : sig
  (*
  We can write down a spec for AddableCore.  The expectation is that any module
  that implements AddableCore must satisfy the following properties:
  *)

  val plus_is_commutative : C.t -> C.t -> bool
  (* a + b = b + a *)

  val plus_is_associative : C.t -> C.t -> C.t -> bool
  (* (a + b) + c = a + (b + c) *)

  val minus_is_plus_inverse : C.t -> bool
  (* a + (-a) = zero *)

  val zero_is_plus_identity : C.t -> bool
  (* a + zero = a *)

  (*
  AddableSpec(C)(E) can be used with TEST_MODULE to run all of these tests
  on all possible combinations of E.examples.  For example:

  TEST_MODULE = AddableSpec(Int)(struct let examples = [1; 2; 3] end)

  will test the 4 properties above with all combinations of the numbers 1, 2,
  and 3.  See impls.ml for many examples.
  *)
end


(******************************************************************************)
(** Multiplication                                                            *)
(******************************************************************************)

(** We follow the same design pattern with multiplication.
    Multiplication's spec can best be defined with some notion of addition, so
    we include the Addable signatures in the corresponding
    Multipliable signatures.

    Ints are multipliable, but vectors are not

    As a design decision, we've made matrices not multipliable, because our
    spec says multiplication is commutative.  We could factor further by
    creating separate module types.  This is part of the full MultipliableSpec,
    defined below.
*)

module type MultipliableCore = sig
  type t

  (* we need ( + ) to define spec for ( * ) *)
  include AddableCore with type t := t

  val ( * ) : t -> t -> t
  val one   : t
end

(* Multipliable spec.  As with AddableSpec, tests can be run by using
   TEST_MODULE(C)(E) *)

module MultipliableSpec (C : MultipliableCore)
                        (E : Examples with type t := C.t) : sig
  (* another nice piece of OCaml syntax: "module type of M" is what you'd expect *)
  include module type of AddableSpec(C)(E)

  val times_is_associative : C.t -> C.t -> C.t -> bool
  (* (a * b) * c = a * (b * c) *)

  val times_is_commutative : C.t -> C.t -> bool
  (* a * b = b * a *)

  val times_distributes_over_plus : C.t -> C.t -> C.t -> bool
  (* a * (b + c) = a * b + a * c *)

  val one_is_times_identity : C.t -> bool
  (* a * one = a *)
end


(******************************************************************************)
(** Comparison and ordering                                                   *)
(******************************************************************************)

(** Same design pattern a third time; we add ordering.  Ints are ordered, but
    complex numbers are not. *)

module type OrderedCore = sig
  include MultipliableCore

  val isPositive : t -> bool
end

(* Ordered spec *)

module OrderedSpec (C : OrderedCore)
                   (E : Examples with type t := C.t) : sig
  include module type of MultipliableSpec(C)(E)

  val sum_of_positives_is_positive : C.t -> C.t -> bool
  (* if isPositive a and isPositive b then isPositive (a + b) *)

  val product_of_positives_is_positive : C.t -> C.t -> bool
  (* if isPositive a and isPositive b then isPositive (a * b) *)

  val everything_is_either_positive_negative_or_zero : C.t -> bool
  (* isPositive a or isPositive (-a) or a = 0 *)
end

(******************************************************************************)
(** Division                                                                  *)
(******************************************************************************)

(* Fourth verse, same as the first.  Only with division
   Complex numbers are dividable, but Ints are not
 *)

module type DividableCore = sig
  include MultipliableCore

  val inv : t -> t
end

module DividableSpec (C : DividableCore)
                     (E : Examples with type t := C.t) : sig
  include module type of MultipliableSpec(C)(E)

  val inv_is_times_inverse : C.t -> bool
  (* if a <> zero then a * inv a = one *)
end

(******************************************************************************)
(** Combining everything                                                      *)
(******************************************************************************)

(* For our application, we'll define a "number" as something that is ordered and
   dividable, and something we can approximate with a float *)

module type NumberCore = sig
  type t

  include DividableCore with type t := t
  include OrderedCore   with type t := t
end

module NumberSpec (C : NumberCore)
                  (E : Examples with type t := C.t) : sig
  include module type of OrderedSpec(C)(E)
  include module type of DividableSpec(C)(E)
end

(******************************************************************************)
(** Extensions ****************************************************************)
(******************************************************************************)

(** We now use functors to define a large number of helper functions that are
    useful for any implementation of the above interfaces.  Each module type
    defined above has a corresponding "Extension" interface, and a "Extend"
    functor that gives the generic implementation.  *)

(** The extra functions that users of an Addable can use *)
module type AddableExtension = sig
  type t
  val (-)    : t -> t -> t
  val (<>)   : t -> t -> bool
end

(** This is the "public" interface: it includes the core and extension *)
module type Addable = sig
  type t

  include AddableCore      with type t := t
  include AddableExtension with type t := t
end

(** A Generic implementation of the Addable interface *)
module ExtendAddable (C : AddableCore) : Addable with type t := C.t


(* the extra functions that users of a Multipliable can use *)
module type MultipliableExtension = sig
  include AddableExtension
  val from_int : int -> t
end

(* the combined user interface for multipliable *)
module type Multipliable = sig
  type t
  include MultipliableCore      with type t := t
  include MultipliableExtension with type t := t
end

module ExtendMultipliable (C : MultipliableCore) : Multipliable with type t := C.t


(* the extra functions that a user of Ordered can use *)
module type OrderedExtension = sig
  include MultipliableExtension

  (** We include some interfaces from the standard library allowing us to put
      Ordered.ts in a Map or a Set *)
  include Set.OrderedType with type t := t
  include Map.OrderedType with type t := t

  val (<)  : t -> t -> bool
  val (>)  : t -> t -> bool
  val (<=) : t -> t -> bool
  val (>=) : t -> t -> bool
end

(* the combined Ordered user interface *)
module type Ordered = sig
  include OrderedCore
  include OrderedExtension with type t := t
end

module ExtendOrdered (C : OrderedCore) : Ordered with type t := C.t



(* the extra functions for users of Dividable types *)
module type DividableExtension = sig
  include MultipliableExtension

  val (/) : t -> t -> t
end

(* the combined Dividable interface *)
module type Dividable = sig
  include DividableCore
  include DividableExtension with type t := t
end

module ExtendDividable (C : DividableCore) : Dividable with type t := C.t


(** Number's extension just combines Dividable and Ordered *)
module type NumberExtension = sig
  include DividableExtension
  include OrderedExtension   with type t := t
end

module type Number = sig
  include NumberCore
  include NumberExtension with type t := t
end

module ExtendNumber (C : NumberCore) : Number with type t := C.t

